import logging
import logging.config

log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "class": "colorlog.ColoredFormatter",
            "format": ("{green}[{asctime}]{reset} {bg_green}{levelname}{reset} {blue}{name}:{reset} {message}"),
            "style": "{",
        },
    },
    "filters": {},
    "handlers": {
        "stdout": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "null": {
            "level": "DEBUG",
            "class": "logging.NullHandler",
        },
    },
    "loggers": {
        "": {
            "handlers": ["stdout"],
            "level": logging.DEBUG,
        },
        "uvicorn": {
            "handlers": ["null"],
            "propagate": False,
        },
        "asyncio": {
            "handlers": ["null"],
            "propagate": False,
        },
    },
}


def setup():
    logging.config.dictConfig(log_config)



from __future__ import annotations

import asyncio
import logging
import typing as t
from abc import ABC
from dataclasses import dataclass

from . import _utils

logger = logging.getLogger(__name__)


type HTML = str


# 1. Bus


class _MESSAGE(ABC):
    type Name = str

    _name = property(lambda self: self.__class__.__name__)
    _data = property(lambda self: self.__dict__)


@dataclass
class SIGNAL(_MESSAGE):
    type T = type[SIGNAL]


@dataclass
class EVERY_SECOND(SIGNAL): ...


@dataclass
class COMMAND(_MESSAGE):
    type T = type[COMMAND]


get_signals_map = lambda: {s.__name__: s for s in SIGNAL.__subclasses__()}


async def emit_signal(sig: SIGNAL | SIGNAL.T) -> None:
    if not isinstance(sig, SIGNAL): sig = sig()

    conn = get_connection()
    logger.info(f'[{conn.id}] -> {sig._name}  {sig._data}')

    for callback in _callbacks.get(sig._name, set()):
        await callback(sig)


from .server import get_connection


async def send_command(cmd: COMMAND | type[COMMAND]) -> None:
    if not isinstance(cmd, COMMAND): cmd = cmd()

    conn = get_connection()
    logger.info(f'[{conn.id}] <- {cmd._name}  {cmd._data}')

    await conn.send_command(cmd)


# 2. Callbacks


type _FunctionCallback = t.Callable[[object, SIGNAL], None]
type _ClassCallback = t.Callable[[SIGNAL], None]

type ClassCallback = t.Awaitable[_ClassCallback]
type FunctionCallback = t.Awaitable[_FunctionCallback]

type AnyCallback = FunctionCallback | ClassCallback
type Callback = FunctionCallback


_callbacks: dict[SIGNAL.Name, set[Callback]] = {}


def subscribe(signal_cls: SIGNAL.T, callback: Callback) -> None:
    sig_name = signal_cls.__name__
    if sig_name not in _callbacks:
        _callbacks[sig_name] = set()

    _callbacks[sig_name].add(callback)


def unsubscribe(signal_cls: SIGNAL.T, callback: Callback) -> None:
    sig_name = signal_cls.__name__
    _callbacks[sig_name].remove(callback)


def on(signal_cls: SIGNAL.T) -> AnyCallback:
    def wrapper(func: AnyCallback) -> AnyCallback:
        self = _utils.get_f_self(func)
        cls_name = _utils.get_f_cls_name(func)

        if self is not None and cls_name:
            # `on` called for `self.method` -> valid callback
            subscribe(signal_cls, func)

        elif self is None and not cls_name:
            # `on` called for module function -> valid callback
            subscribe(signal_cls, func)

        elif self is None and cls_name:
            # `on` called for class function -> not valid callback...
            # only valid in component interface
            from .layout import Component
            Component.schedule_callback(signal_cls, cls_name, func.__name__)

        elif self is not None and not cls_name:
            raise RuntimeError  # no way...

        return func

    return wrapper


# 3. App


from .layout import Component
from .layout import Layout
from .server import Server


class App:
    def run(self, **params: dict) -> None:
        try:
            asyncio.run(self.task(**params))
        except KeyboardInterrupt:
            pass

    async def task(
        self,
        layout: t.Iterable[Component | HTML] = [],
        title: str = 'Sundash',
    ) -> None:
        self.layout = Layout()
        for comp in layout: self.layout.append(comp)

        self.server = Server(html_title=title)
        await self.server.task()



import functools
import typing as t
from dataclasses import dataclass
from dataclasses import is_dataclass

from .core import COMMAND
from .core import HTML
from .core import SIGNAL
from .core import Callback
from .core import on
from .core import send_command
from .core import subscribe
from .core import unsubscribe
from .server import CLIENT_CONNECTED
from .server import CLIENT_DISCONNECTED
from .server import get_connection


class Var[T]:
    type Key = str
    type Value = T


type VarStorage = dict[Var.Key, Var.Value]


@dataclass
class LAYOUT_UPDATED(SIGNAL): ...


@dataclass
class LAYOUT_CLEAN(SIGNAL): ...


@dataclass
class VAR_SET(SIGNAL): ...


@dataclass
class BUTTON_CLICK(SIGNAL):
    button_id: str


@dataclass
class UPDATE_LAYOUT(COMMAND):
    html: HTML
    vars: VarStorage


@dataclass
class CLEAR_LAYOUT(COMMAND): ...


@dataclass
class SET_VAR(COMMAND):
    key: Var.Key
    value: Var.Value


class Component:
    html: HTML = ''

    @dataclass
    class Vars: ...

    _callbacks: set[tuple[SIGNAL.T, str, str]] = set()

    def __init__(self):
        if not is_dataclass(self.Vars):
            raise RuntimeError

        # TODO: init procedural values
        self.vars: dict = self.Vars().__dict__
        self.conn_id = get_connection().id

    def callback_wrapper(self, callback: Callback) -> Callback:
        async def wrapper(sig: SIGNAL):
            conn_id = get_connection().id
            if conn_id != self.conn_id:
                return
            await callback(sig)

        return functools.wraps(callback)(wrapper)

    @classmethod
    def schedule_callback(
        cls, signal_cls: SIGNAL.T, cls_name: str, func_name: str
    ) -> None:
        cls._callbacks.add((signal_cls, cls_name, func_name))

    def callbacks_map(self) -> t.Generator:
        cls = self.__class__
        for signal_cls, cls_name, func_name in self.__class__._callbacks:
            if cls.__name__ != cls_name: continue

            callback = getattr(self, func_name)
            yield signal_cls, callback

    def subscribe_callbacks(self) -> None:
        for signal_cls, callback in self.callbacks_map():
            subscribe(signal_cls, self.callback_wrapper(callback))

    def unsubscribe_callbacks(self) -> None:
        for signal_cls, callback in self.callbacks_map():
            unsubscribe(signal_cls, callback)

    async def set(self, key: Var.Key, value: Var.Value) -> None:
        self.vars[key] = value
        await send_command(SET_VAR(key=key, value=value))


class Layout(list[type[Component]]):
    _sessions: dict[int, list[Component]] = {}

    def __init__(self):
        super().__init__()

        on(CLIENT_CONNECTED)(self.open_session)
        on(CLIENT_DISCONNECTED)(self.close_session)

    def append(self, item: type[Component] | HTML) -> None:
        if isinstance(item, str):
            class HTMLComponent(Component):
                html = item

            comp = HTMLComponent
        else:
            comp = item

        super().append(comp)

    async def open_session(self, _) -> None:
        layout = []
        for comp_cls in self:
            comp = comp_cls()
            comp.subscribe_callbacks()
            layout.append(comp)

        data = {}
        for comp in layout:
            data.update(**comp.vars)

        html = ''.join((c.html for c in layout))

        id = get_connection().id
        self._sessions[id] = layout

        await send_command(UPDATE_LAYOUT(html=html, vars=data))

    async def close_session(self, _) -> None:
        id = get_connection().id
        for comp in self._sessions[id]:
            comp.unsubscribe_callbacks()

        self._sessions.pop(id)



import contextvars
import json
import logging
import subprocess
from dataclasses import dataclass

import uvicorn
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.staticfiles import StaticFiles
from starlette.types import Receive
from starlette.types import Scope
from starlette.types import Send
from starlette.websockets import WebSocket
from starlette.websockets import WebSocketDisconnect

from .core import COMMAND
from .core import HTML
from .core import SIGNAL
from .core import emit_signal
from .core import get_signals_map
from .logging import log_config

logger = logging.getLogger(__name__)


@dataclass
class CLIENT_CONNECTED(SIGNAL): ...


@dataclass
class CLIENT_DISCONNECTED(SIGNAL): ...


class WSConnection:
    __id = 0

    @classmethod
    def new_id(cls) -> int:
        cls.__id += 1
        return cls.__id

    def __init__(self, socket: WebSocket) -> None:
        self.id = self.__class__.new_id()
        self.socket = socket

    async def receive_signal(self) -> None:
        message = await self.socket.receive_text()

        signal_name, data = message.split(" ", 1)
        signal_cls = get_signals_map()[signal_name]
        data = json.loads(data)

        await emit_signal(signal_cls(**data))

    async def send_command(self, cmd: COMMAND) -> None:
        cmd_name = cmd.__class__.__name__
        cmd_params = json.dumps(cmd.__dict__)
        await self.socket.send_text(f'{cmd_name} {cmd_params}')


_conn = contextvars.ContextVar('_conn', default=None)


def set_connection(conn: WSConnection) -> None:
    _conn.set(conn)


def get_connection() -> WSConnection:
    return _conn.get()


class Server:
    _EXIT_CODE = 1
    ALLOWED_STATIC_FILES = ('.html', 'css', '.js', '.map', '.ico')

    class _ASGIServer(uvicorn.Server):
        def install_signal_handlers(self) -> None:
            # replace default signal catch
            # because I want `Ctrl + C` to work correct
            pass

    def __init__(
        self,
        host: str = 'localhost',
        port: int = 5000,
        html_title: str = 'Sundash',
    ):
        self.host = host
        self.port = port
        self.html_title = html_title

        self._index_html: str = None
        self._connections: dict[int: WSConnection] = {}

    async def task(self) -> None:
        build_ui()
        logger.info(f'Starting server at http://{self.host}:{self.port}/')

        config = uvicorn.Config(
            app=self,
            host=self.host,
            port=self.port,
            log_level='debug',
            log_config=log_config,
        )
        server = self._ASGIServer(config=config)
        try:
            await server.serve()
        finally:
            logger.info('Shutting down server')

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        if scope['type'] == 'lifespan':
            exit_code = await self._handle_lifespan(scope, receive, send)
            if exit_code:
                return

        elif scope['type'] == 'http':
            await self._handle_http_request(scope, receive, send)

        elif scope['type'] == 'websocket':
            await self._handle_websocket_request(scope, receive, send)

        else:
            response = HTMLResponse(
                content='<b>Not allowed</b>', status_code=405
            )
            await response(scope, receive, send)

    async def _handle_lifespan(
        self, scope: Scope, receive: Receive, send: Send
    ) -> int | None:
        message = await receive()

        if message['type'] == 'lifespan.startup':
            await send({'type': 'lifespan.startup.complete'})
            return None

        elif message['type'] == 'lifespan.shutdown':
            await send({'type': 'lifespan.shutdown.complete'})
            return self._EXIT_CODE

        else:
            return None

    async def _handle_http_request(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:

        request = Request(scope, receive, send)
        path = request.url.components.path

        if any([path.endswith(ext) for ext in self.ALLOWED_STATIC_FILES]):
            static_files = StaticFiles(directory='static')
            await static_files(scope, receive, send)

        else:
            content = self._get_index_html_content()
            response = HTMLResponse(content=content, status_code=200)
            await response(scope, receive, send)

    def _get_index_html_content(self) -> HTML:
        if not self._index_html:
            with open('static/index.html') as f:
                self._index_html = (
                    f.read().replace('{{ _app_title }}', self.html_title)
                )
        return self._index_html

    async def _handle_websocket_request(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        socket = WebSocket(scope=scope, receive=receive, send=send)
        await socket.accept()

        conn = WSConnection(socket)
        set_connection(conn)

        try:
            await emit_signal(CLIENT_CONNECTED)
            while True:
                await conn.receive_signal()

        except WebSocketDisconnect:
            pass

        except Exception as exc:
            logger.exception(exc)

        finally:
            await emit_signal(CLIENT_DISCONNECTED)
            set_connection(None)


def build_ui():
    logger.info('Building web UI...')
    # FIXME control output and redirect to logger
    subprocess.run(['npm', 'run', 'build'])
